package com.reportmill.webobjects;
import com.webobjects.foundation.*;
import com.webobjects.eocontrol.*;
import com.reportmill.foundation.*;
import java.lang.reflect.*;
import java.util.*;

/************************************* - All files should be 120 chars wide - ******************************************

    RMWebObjects
    
    This class compensates for WebObjects use of non-standard collection and core classes (NSDictionary, NSArray, NSTimeStamp) and for EOCustomObject, which is a hybrid of a Map-style/custom-object-style object.
    
    This class installs an instance of itselfs in the ReportMill class's 'appServer' static variable, so that it's 'convertFromAppServerType()' method will get called for all evaluated keys. This method returns a wrapper object for each of the above mentioned classes, which simply provide standard access with little over-head.
    
    In addition, this class implements a writeObject() method that uses a hook in RMXMLWriter to also call convertFromAppServerType. When writeObject() is called, convertFromAppServerType() actually returns standard Java collection classes, since XML writer needs to do more than simple data access. The over-head doesn't matter, since XML writing is only for development purposes.

***********************************************************************************************************************/
public class RMWebObjects implements ReportMill.AppServer {
    static RMXMLWriter    _writer = null;
    static RMWebObjects   _shared = new RMWebObjects();

// Install instance of this class before any RMWebObjects functionality is used
public static void setAppServer() { ReportMill.appServer = _shared; }
static { setAppServer(); }

// Type conversion
public Object convertFromAppServerType(Object obj)
{    
    // Handle EOCustomObject (turn into Map if doing writeObject, else RMEOWrapper for reporting)
    if(obj instanceof EOCustomObject) {
        if(_writer!=null) {
            Map map = ((EOCustomObject)obj).snapshot().hashtable();
            if(obj instanceof EOGenericRecord)
                map.put("Class", ((EOGenericRecord)obj).classDescription().entityName());
            else map.put("Class", obj.getClass().getName());
            addMembers(obj, map);
            obj = map;
        }
        else obj = new RMEOWrapper((EOCustomObject)obj);
    }
    
    // Handle NSDictionary (turn into Hashtable if doing writeObject or RMNSDictionaryWrapper for reporting)
    else if(obj instanceof NSDictionary) {
        if(_writer!=null)
            obj = ((NSDictionary)obj).hashtable();
        else obj = new RMNSDictionaryWrapper((NSDictionary)obj);
    }
    
    // Handle NSArray (wrap in RMNSArrayWrapper which implements List interface)
    else if(obj instanceof NSArray)
        obj = new RMNSArrayWrapper((NSArray)obj);
        
    // Handle NSData (turn it into a byte array)
    else if(obj instanceof NSData)
        obj = ((NSData)obj).bytes();
    
    // Handle NSKeyValueCoding.Null (substitute null)
    else if(obj instanceof NSKeyValueCoding.Null)
        return null;
     
    // Handle NSTimestamp (demote to java.util.Date if doing writeObject)
    else if(obj instanceof NSTimestamp && _writer!=null)
        obj = new Date(((NSTimestamp)obj).getTime());

    return obj;
}

public static void writeObject(Object obj, String aPath) { writeObject(obj, null, aPath, 3); }
public static void writeObject(Object obj, String aPath, int dos) { writeObject(obj, null, aPath, dos); }
public static void writeObject(Object obj, Object userInfo, String aPath) { writeObject(obj, userInfo, aPath, 3); }
public static void writeObject(Object object, Object userInfo, String aPath, int degreeOfSeparation)
{
    // Get new RMWOXMLWriter and writeObject (and then clear _writer)
    setAppServer();
    _writer = new RMWOXMLWriter();
    _writer.writeObject(object, userInfo, aPath, degreeOfSeparation);
    _writer = null;
}

// An RMXMLWriter subclass to perform WO conversions and ignore uneccessary WO/EOF fields
public static class RMWOXMLWriter extends RMXMLWriter {
    public RMWOXMLWriter() {
        ignoreClass("com.webobjects.eocontrol.EOEditingContext");
        ignoreMember("allPropertyKeys");
        ignoreMember("attributeKeys");
        ignoreMember("Class");
        ignoreMember("classDescription");
        ignoreMember("classDescriptionForInstances");
        ignoreMember("editingContext");
        ignoreMember("entityName");
        ignoreMember("faultHandler");
        ignoreMember("globalID");
        ignoreMember("isFault");
        ignoreMember("KeyPathSeparator");
        ignoreMember("NullValue");
        ignoreMember("opaqueState");
        ignoreMember("snapshot");
        ignoreMember("toOneRelationshipKeys");
        ignoreMember("toManyRelationshipKeys");
        ignoreMember("userPresentableDescription");
    }
    public boolean ignore(String aName) { return aName.startsWith("_") || aName.startsWith("eo") || super.ignore(aName); }
}

// An inner class to wrap EOCustomObject
public static class RMEOWrapper implements RMKey.ValueForKey {
    EOCustomObject _eog;
    public RMEOWrapper(EOCustomObject eog) { _eog = eog; }
    public Object valueForKey(String aKey) {
        try { return _eog.valueForKey(aKey); }
        catch(Exception e) { return null; }
    }
    public boolean equals(Object anObj) {
        if(anObj instanceof RMEOWrapper) anObj = ((RMEOWrapper)anObj)._eog;
        return _eog.equals(anObj);
    }
    public int hashCode() { return _eog.hashCode(); }
}

// An inner class to wrap NSDictionary
public static class RMNSDictionaryWrapper implements RMKey.ValueForKey {
    NSDictionary _dict;
    public RMNSDictionaryWrapper(NSDictionary aDict) { _dict = aDict; }
    public Object valueForKey(String aKey) {
        try { return _dict.valueForKey(aKey); }
        catch(Exception e) { return null; }
    }
    public boolean equals(Object anObj) {
        if(anObj instanceof RMNSDictionaryWrapper) anObj = ((RMNSDictionaryWrapper)anObj)._dict;
        return _dict.equals(anObj);
    }
    public int hashCode() { return _dict.hashCode(); }
}

// An inner class to wrap NSArray
public static class RMNSArrayWrapper extends AbstractList {
    NSArray _array;
    public RMNSArrayWrapper(NSArray anArray) { _array = anArray; }
    public int size() { return _array.count(); }
    public Object get(int anIndex) { return _array.objectAtIndex(anIndex); }
    public boolean equals(Object anObj) {
        if(anObj instanceof RMNSArrayWrapper) anObj = ((RMNSArrayWrapper)anObj)._array;
        return _array.equals(anObj);
    }
    public int hashCode() { return _array.hashCode(); }
}

public static void addMembers(Object anObj, Map aMap)
{
    // Load fields into map
    Class objClass = anObj.getClass();
    Field fields[] = objClass.getFields();
    for(int i=0; i<fields.length; i++) {
        Field field = fields[i];
        String fieldName = field.getName();
        
        // If field should be ignored, just continue
        if(_writer.ignore(fieldName))
            continue;
        
        // Get value for field and add to aMap if not null
        try {
            Object obj = field.get(anObj);
            if(obj!=null) aMap.put(fieldName, obj);
        }
        catch(Exception e) { }
    }
    
    // Load methods into map
    Method methods[] = objClass.getMethods();
    for(int i=0; i<methods.length; i++) {
        Method method = methods[i];
        String methodName = method.getName();
        
        // If method should be ignored or has no return value or is static or not public, just continue
        int modifiers = method.getModifiers();
        if(_writer.ignore(methodName) || method.getReturnType()==void.class)
            continue;
        if(Modifier.isStatic(modifiers) || !Modifier.isPublic(modifiers))
            continue;
        if(method.getParameterTypes().length>0)
            continue;
            
        // Get value for method and add to aMap if not null
        System.out.println("Accessing: " + anObj.getClass().getName() + "." + methodName);
        Object obj = null; try { obj = method.invoke(anObj, new Object[0]); } catch(Exception e) { }
        if(obj!=null)
            aMap.put(methodName, obj);
    }
}

}